use std::{collections::HashMap, str::FromStr};

use axum::{
    extract::Path,
    http::{HeaderName, HeaderValue, Uri},
    response::{Html, IntoResponse, Response},
};
use networkd::{ServerApp, ServerAppProps, Route};
use tower_http::
    trace::{DefaultMakeSpan, DefaultOnRequest, DefaultOnResponse}
;
use tracing::Level;

use yew::ServerRenderer;
use yew_router::Routable;

static INDEX_PAGE: &str = include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/dist/index.html"));
static DIST_DIR: include_dir::Dir = include_dir::include_dir!("$CARGO_MANIFEST_DIR/dist");

#[tracing::instrument]
async fn rendering_pages(
    url: Uri,
    axum::extract::Query(queries): axum::extract::Query<HashMap<String, String>>,
) -> Response {
    let url = url.to_string();
    let url_ = url.clone();
    let renderer = ServerRenderer::<ServerApp>::with_props(move || ServerAppProps {
        url: url_.into(),
        queries,
    });

    let route = Route::recognize(&url);
    let status_code = match route {
        None | Some(Route::NotFound) => axum::http::StatusCode::NOT_FOUND,
        _ => axum::http::StatusCode::OK,
    };
    
    let rendered = renderer.render().await;
    (status_code, Html::from(INDEX_PAGE.replace("<!--inject-->", &rendered))).into_response()
}

#[tracing::instrument]
async fn static_assets(Path(path): Path<String>) -> Response {
    let Some(file) = DIST_DIR.get_file(&path) else {
        return axum::http::StatusCode::NOT_FOUND.into_response();
    };
    let guess = mime_guess::from_path(&path)
        .first()
        .unwrap_or(mime::APPLICATION_OCTET_STREAM)
        .to_string();

    let headers = axum::http::HeaderMap::from_iter([(
        HeaderName::from_str("Content-Type").unwrap(),
        HeaderValue::from_str(&guess).unwrap(),
    )]);
    (headers, file.contents()).into_response()
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    #[cfg(feature = "console-subscriber")]
    tracing_subscriber::registry()
        .with(console_subscriber::spawn())
        .with(tracing_subscriber::fmt::layer())
        .init();

    #[cfg(not(feature = "console-subscriber"))]
    tracing_subscriber::fmt().init();

    let api_router = axum::Router::new();
    let oauth_router = axum::Router::new();

    let router = axum::Router::new()
        // not required to login
        .route("/static/*path", axum::routing::get(static_assets))
        // login / logout
        .route("/login", axum::routing::get(rendering_pages))
        .route("/logout", axum::routing::get(rendering_pages))
        // with login
        .route("/networks", axum::routing::get(rendering_pages))
        // nesting router
        .nest("/api/", api_router)
        .nest("/o/", oauth_router)

        // fallback
        .fallback(rendering_pages)
        .layer(
            tower::ServiceBuilder::new().layer(
                tower_http::trace::TraceLayer::new_for_http()
                    .make_span_with(DefaultMakeSpan::new().level(Level::INFO))
                    .on_request(DefaultOnRequest::new().level(Level::INFO))
                    .on_response(DefaultOnResponse::new().level(Level::INFO)),
            ),
        );

    let listener = tokio::net::TcpListener::bind("0.0.0.0:1234").await?;
    axum::serve(listener, router).await?;

    Ok(())
}
